@@ -1,6 +1,6 @@
-/* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2018 Brazil
+  Copyright(C) 2009-2018  Brazil
+  Copyright(C) 2020  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -40,22 +40,31 @@ typedef unsigned int msgpack_size_t;
   msgpack_pack_raw_body(packer, value, size)
 
 #  define MSGPACK_OBJECT_STR MSGPACK_OBJECT_RAW
-#  define MSGPACK_OBJECT_FLOAT MSGPACK_OBJECT_DOUBLE
+#  define MSGPACK_OBJECT_FLOAT64 MSGPACK_OBJECT_DOUBLE
 
 #  define MSGPACK_OBJECT_STR_PTR(object)  (object)->via.raw.ptr
 #  define MSGPACK_OBJECT_STR_SIZE(object) (object)->via.raw.size
 
-#  define MSGPACK_OBJECT_FLOAT_VALUE(object) (object)->via.dec
+#  define MSGPACK_OBJECT_FLOAT64_VALUE(object) (object)->via.dec
 
 #  define MSGPACK_UNPACKER_NEXT(unpacker, unpacked)     \
   msgpack_unpacker_next((unpacker), (unpacked))
 # else /* MSGPACK_VERSION_MAJOR < 1 */
 typedef size_t msgpack_size_t;
 
+#  if MSGPACK_VERSION_MAJOR == 1
+#   define MSGPACK_OBJECT_FLOAT64 MSGPACK_OBJECT_FLOAT
+#  elif MSGPACK_VERSION_MAJOR == 2 && MSGPACK_VERSION_MINOR < 1
+#   define MSGPACK_OBJECT_FLOAT64 MSGPACK_OBJECT_DOUBLE
+#  else
+#   define MSGPACK_HAVE_FLOAT32
+#  endif
+
 #  define MSGPACK_OBJECT_STR_PTR(object)  (object)->via.str.ptr
 #  define MSGPACK_OBJECT_STR_SIZE(object) (object)->via.str.size
 
-#  define MSGPACK_OBJECT_FLOAT_VALUE(object) (object)->via.f64
+#  define MSGPACK_OBJECT_FLOAT32_VALUE(object) (object)->via.f64
+#  define MSGPACK_OBJECT_FLOAT64_VALUE(object) (object)->via.f64
 
 #  define MSGPACK_UNPACKER_NEXT(unpacker, unpacked)                     \
   msgpack_unpacker_next((unpacker), (unpacked)) == MSGPACK_UNPACK_SUCCESS
