@@ -1,6 +1,6 @@
 /*
   Copyright(C) 2009-2018  Brazil
-  Copyright(C) 2018-2020  Sutou Kouhei <kou@clear-code.com>
+  Copyright(C) 2018-2021  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -808,7 +808,7 @@ dump_table(grn_ctx *ctx, grn_dumper *dumper, grn_obj *table)
     grn_obj sub_output;
     GRN_TEXT_PUTS(ctx, dumper->output, " --normalizer ");
     GRN_TEXT_INIT(&sub_output, 0);
-    grn_table_get_normalizer_string(ctx, table, &sub_output);
+    grn_table_get_normalizers_string(ctx, table, &sub_output);
     dump_string(ctx, dumper, &sub_output);
     GRN_OBJ_FIN(ctx, &sub_output);
   }
