@@ -16,6 +16,7 @@
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
+#include "../grn_ctx.hpp"
 #include "../grn_expr.h"
 #include "../grn_proc.h"
 #include "../grn_hash.h"
@@ -94,13 +95,15 @@ namespace {
       op_(op),
       selector_data_(selector_data),
       tag_(tag),
-      match_columns_string_(nullptr),
+      match_columns_raw_(nullptr),
       query_expander_name_(nullptr),
       query_options_(nullptr),
       default_mode_(GRN_OP_MATCH),
       default_operator_(GRN_OP_AND),
       flags_(GRN_EXPR_SYNTAX_QUERY),
       flags_specified_(-1),
+      enough_filtered_ratio_(-1),
+      max_n_enough_filtered_records_(-1),
       match_columns_(nullptr) {
     }
 
@@ -153,15 +156,46 @@ namespace {
       }
     }
 
+    void
+    init_table_selector(grn_ctx *ctx,
+                        grn_table_selector *table_selector,
+                        grn_obj *condition,
+                        grn_operator op,
+                        grn_id min_id) {
+      grn_table_selector_init(ctx,
+                              table_selector,
+                              table_,
+                              condition,
+                              op);
+      grn_table_selector_set_min_id(ctx,
+                                    table_selector,
+                                    min_id);
+      grn_table_selector_set_weight_factor(ctx,
+                                           table_selector,
+                                           get_weight_factor());
+      if (enough_filtered_ratio_ >= 0) {
+        grn_table_selector_set_enough_filtered_ratio(ctx_,
+                                                     table_selector,
+                                                     enough_filtered_ratio_);
+      }
+      if (max_n_enough_filtered_records_ >= 0) {
+        grn_table_selector_set_max_n_enough_filtered_records(
+          ctx_,
+          table_selector,
+          max_n_enough_filtered_records_);
+      }
+    }
+
     bool
     parse_match_columns_arg() {
-      match_columns_string_ = args_[0];
-      if (!grn_obj_is_text_family_bulk(ctx_, match_columns_string_)) {
+      match_columns_raw_ = args_[0];
+      if (!grn_obj_is_text_family_bulk(ctx_, match_columns_raw_) &&
+          !grn_obj_is_vector(ctx_, match_columns_raw_)) {
         grn::TextBulk inspected(ctx_);
-        grn_inspect(ctx_, *inspected, match_columns_string_);
+        grn_inspect(ctx_, *inspected, match_columns_raw_);
         GRN_PLUGIN_ERROR(ctx_,
                          GRN_INVALID_ARGUMENT,
-                         "%s 1st argument must be string: <%.*s>",
+                         "%s 1st argument must be string or array: <%.*s>",
                          tag_,
                          (int)GRN_TEXT_LEN(*inspected),
                          GRN_TEXT_VALUE(*inspected));
@@ -188,7 +222,13 @@ namespace {
         &flags_specified_,                                              \
         "options",                                                      \
         GRN_PROC_OPTION_VALUE_RAW,                                      \
-        &query_options_ptr
+        &query_options_ptr,                                             \
+        "enough_filtered_ratio",                                        \
+        GRN_PROC_OPTION_VALUE_DOUBLE,                                   \
+        &enough_filtered_ratio_,                                        \
+        "max_n_enough_filtered_records",                                \
+        GRN_PROC_OPTION_VALUE_INT64,                                    \
+        &max_n_enough_filtered_records_
       if (selector_data_) {
         grn_selector_data_parse_options(ctx_,
                                         selector_data_,
@@ -237,23 +277,34 @@ namespace {
 
     bool
     prepare_match_columns() {
-      if (GRN_TEXT_LEN(match_columns_string_) == 0) {
-        return true;
+      if (grn_obj_is_text_family_bulk(ctx_, match_columns_raw_)) {
+        return prepare_match_columns_bulk();
+      } else {
+        return prepare_match_columns_vector();
       }
+    }
 
-      grn_obj *dummy_variable;
-      GRN_EXPR_CREATE_FOR_QUERY(ctx_,
-                                table_,
-                                match_columns_,
-                                dummy_variable);
-      if (!match_columns_) {
-        return false;
+    bool
+    prepare_match_columns_one(const char *match_columns_string,
+                              size_t match_columns_string_length,
+                              grn_obj **match_columns_inout) {
+      grn_obj *match_columns = *match_columns_inout;
+      if (!match_columns) {
+        grn_obj *dummy_variable;
+        GRN_EXPR_CREATE_FOR_QUERY(ctx_,
+                                  table_,
+                                  match_columns,
+                                  dummy_variable);
+        *match_columns_inout = match_columns;
+        if (!match_columns) {
+          return false;
+        }
       }
 
       grn_expr_parse(ctx_,
-                     match_columns_,
-                     GRN_TEXT_VALUE(match_columns_string_),
-                     GRN_TEXT_LEN(match_columns_string_),
+                     match_columns,
+                     match_columns_string,
+                     match_columns_string_length,
                      NULL,
                      GRN_OP_MATCH,
                      GRN_OP_AND,
@@ -261,6 +312,50 @@ namespace {
       return ctx_->rc == GRN_SUCCESS;
     }
 
+    virtual bool
+    prepare_match_columns_bulk() {
+      if (GRN_TEXT_LEN(match_columns_raw_) == 0) {
+        return true;
+      }
+
+      return prepare_match_columns_one(GRN_TEXT_VALUE(match_columns_raw_),
+                                       GRN_TEXT_LEN(match_columns_raw_),
+                                       &match_columns_);
+    }
+
+    virtual bool
+    prepare_match_columns_vector() {
+      auto n_match_columns = grn_vector_size(ctx_, match_columns_raw_);
+      if (n_match_columns == 0) {
+        return true;
+      }
+
+      uint32_t n_sub_match_columns = 0;
+      for (unsigned int i = 0; i < n_match_columns; ++i) {
+        const char *sub_match_columns_string;
+        auto sub_match_columns_string_length =
+          grn_vector_get_element(ctx_,
+                                 match_columns_raw_,
+                                 i,
+                                 &sub_match_columns_string,
+                                 NULL,
+                                 NULL);
+        if (sub_match_columns_string_length == 0) {
+          continue;
+        }
+        if (!prepare_match_columns_one(sub_match_columns_string,
+                                       sub_match_columns_string_length,
+                                       &match_columns_)) {
+          return false;
+        }
+        n_sub_match_columns++;
+        if (n_sub_match_columns >= 2) {
+          grn_expr_append_op(ctx_, match_columns_, GRN_OP_OR, 2);
+        }
+      }
+      return true;
+    }
+
     bool
     expand_query(grn_obj *query, grn::TextBulk &expanded_query)
     {
@@ -285,6 +380,67 @@ namespace {
       return ctx_->rc == GRN_SUCCESS;
     }
 
+    grn_obj *
+    build_condition(grn_ctx *ctx,
+                    grn_obj *match_columns,
+                    const char *query,
+                    size_t query_length) {
+      grn_obj *condition;
+      grn_obj *dummy_variable;
+      GRN_EXPR_CREATE_FOR_QUERY(ctx,
+                                table_,
+                                condition,
+                                dummy_variable);
+      if (!condition) {
+        return nullptr;
+      }
+      grn_expr_parse(ctx,
+                     condition,
+                     query,
+                     query_length,
+                     match_columns,
+                     default_mode_,
+                     default_operator_,
+                     flags_);
+      if (ctx->rc != GRN_SUCCESS) {
+        grn_obj_close(ctx, condition);
+        return nullptr;
+      }
+      if (query_options_) {
+        grn_expr_set_query_options(ctx, condition, query_options_);
+      }
+      return condition;
+    }
+
+    bool
+    select_one(grn_ctx *ctx,
+               grn_obj *match_columns,
+               const char *query,
+               size_t query_length,
+               grn_operator op,
+               grn_id min_id,
+               grn_obj **result_set) {
+      auto condition = build_condition(ctx,
+                                       match_columns,
+                                       query,
+                                       query_length);
+      if (!condition) {
+        return false;
+      }
+      grn::UniqueObj unique_condition(ctx, condition);
+      grn_table_selector table_selector;
+      init_table_selector(ctx,
+                          &table_selector,
+                          condition,
+                          op,
+                          min_id);
+      *result_set = grn_table_selector_select(ctx,
+                                              &table_selector,
+                                              *result_set);
+      grn_table_selector_fin(ctx, &table_selector);
+      return ctx_->rc == GRN_SUCCESS;
+    }
+
     grn_ctx *ctx_;
     grn_obj *table_;
     int n_args_;
@@ -294,13 +450,15 @@ namespace {
     grn_selector_data *selector_data_;
     const char *tag_;
 
-    grn_obj *match_columns_string_;
+    grn_obj *match_columns_raw_;
     grn_obj *query_expander_name_;
     grn_obj *query_options_;
     grn_operator default_mode_;
     grn_operator default_operator_;
     grn_expr_flags flags_;
     grn_expr_flags flags_specified_;
+    double enough_filtered_ratio_;
+    int64_t max_n_enough_filtered_records_;
     grn_obj *match_columns_;
   };
 
@@ -321,14 +479,10 @@ namespace {
                         op,
                         selector_data,
                         "[query]"),
-      query_(nullptr),
-      condition_(nullptr) {
+      query_(nullptr) {
     }
 
     ~QueryExecutor() override {
-      if (condition_) {
-        grn_obj_close(ctx_, condition_);
-      }
     }
 
   private:
@@ -377,9 +531,6 @@ namespace {
       if (!prepare_match_columns()) {
         return false;
       }
-      if (!prepare_condition()) {
-        return false;
-      }
       return true;
     }
 
@@ -414,67 +565,26 @@ namespace {
     }
 
     bool
-    prepare_condition() {
+    run() override {
       if (GRN_TEXT_LEN(query_) == 0) {
         return true;
       }
 
-      grn_obj *dummy_variable;
-      GRN_EXPR_CREATE_FOR_QUERY(ctx_,
-                                table_,
-                                condition_,
-                                dummy_variable);
-      if (!condition_) {
-        return false;
-      }
-
       grn::TextBulk expanded_query(ctx_);
       if (!expand_query(query_, expanded_query)) {
         return false;
       }
-      grn_expr_parse(ctx_,
-                     condition_,
-                     GRN_TEXT_VALUE(*expanded_query),
-                     GRN_TEXT_LEN(*expanded_query),
-                     match_columns_,
-                     default_mode_,
-                     default_operator_,
-                     flags_);
-      if (ctx_->rc != GRN_SUCCESS) {
-        return false;
-      }
-      if (query_options_) {
-        grn_expr_set_query_options(ctx_, condition_, query_options_);
-      }
-      return ctx_->rc == GRN_SUCCESS;
-    }
 
-    bool
-    run() override {
-      if (!condition_) {
-        return true;
-      }
-      grn_table_selector table_selector;
-      grn_table_selector_init(ctx_,
-                              &table_selector,
-                              table_,
-                              condition_,
-                              op_);
-      grn_table_selector_set_min_id(ctx_,
-                                    &table_selector,
-                                    get_min_id());
-      grn_table_selector_set_weight_factor(ctx_,
-                                           &table_selector,
-                                           get_weight_factor());
-      grn_table_selector_select(ctx_,
-                                &table_selector,
-                                res_);
-      grn_table_selector_fin(ctx_, &table_selector);
-      return ctx_->rc == GRN_SUCCESS;
+      return select_one(ctx_,
+                        match_columns_,
+                        GRN_TEXT_VALUE(*expanded_query),
+                        GRN_TEXT_LEN(*expanded_query),
+                        op_,
+                        get_min_id(),
+                        &res_);
     }
 
     grn_obj *query_;
-    grn_obj *condition_;
   };
 
   class QueryParallelOrExecutor : public BaseQueryExecutor {
@@ -506,6 +616,51 @@ namespace {
     }
 
   private:
+    bool
+    prepare_match_columns_bulk() override {
+      if (!BaseQueryExecutor::prepare_match_columns_bulk()) {
+        return false;
+      }
+      if (match_columns_) {
+        grn_expr_match_columns_split(ctx_,
+                                     match_columns_,
+                                     &sub_match_columns_);
+      } else {
+        GRN_PTR_PUT(ctx_, &sub_match_columns_, NULL);
+      }
+      return true;
+    }
+
+    bool
+    prepare_match_columns_vector() override {
+      auto n_match_columns = grn_vector_size(ctx_, match_columns_raw_);
+      if (n_match_columns == 0) {
+        return true;
+      }
+
+      for (unsigned int i = 0; i < n_match_columns; ++i) {
+        const char *sub_match_columns_string;
+        auto sub_match_columns_string_length =
+          grn_vector_get_element(ctx_,
+                                 match_columns_raw_,
+                                 i,
+                                 &sub_match_columns_string,
+                                 NULL,
+                                 NULL);
+        if (sub_match_columns_string_length == 0) {
+          continue;
+        }
+        grn_obj *sub_match_columns = nullptr;
+        if (!prepare_match_columns_one(sub_match_columns_string,
+                                       sub_match_columns_string_length,
+                                       &sub_match_columns)) {
+          return false;
+        }
+        GRN_PTR_PUT(ctx_, &sub_match_columns_, sub_match_columns);
+      }
+      return true;
+    }
+
     bool
     prepare() override {
       if (!validate_args()) {
@@ -531,13 +686,6 @@ namespace {
       if (!prepare_match_columns()) {
         return false;
       }
-#ifdef GRN_WITH_APACHE_ARROW
-      if (match_columns_) {
-        grn_expr_match_columns_split(ctx_, match_columns_, &sub_match_columns_);
-      } else {
-        GRN_PTR_PUT(ctx_, &sub_match_columns_, NULL);
-      }
-#endif
       return true;
     }
 
@@ -628,114 +776,133 @@ namespace {
     bool
     select_parallel(std::shared_ptr<::arrow::internal::ThreadPool> pool) {
       grn_id min_id = get_min_id();
+      // merge isn't parallel. This is just used for queue.
+      auto merge_pool_result = ::arrow::internal::ThreadPool::MakeEternal(1);
+      if (!grnarrow::check(ctx_,
+                           merge_pool_result,
+                           "%s failed to create a thread pool for merging",
+                           tag_)) {
+        return false;
+      }
+      auto merge_pool = *merge_pool_result;
       std::mutex mutex;
       std::vector<grn::UniqueObj> unique_sub_results;
+      auto merge = [&](grn_obj *sub_result) {
+        std::lock_guard<std::mutex> lock(mutex);
+        if (unique_sub_results.empty()) {
+          unique_sub_results.emplace_back(ctx_, sub_result);
+        } else {
+          if (ctx_->rc == GRN_SUCCESS) {
+            grn_table_setoperation(ctx_,
+                                   unique_sub_results[0].get(),
+                                   sub_result,
+                                   unique_sub_results[0].get(),
+                                   GRN_OP_OR);
+          }
+          unique_sub_results.emplace_back(ctx_, sub_result);
+        }
+        return arrow::Status::OK();
+      };
+      std::vector<::arrow::Future<>> merge_futures;
       auto select = [&](grn_obj *match_columns, const std::string &query) {
         auto sub_ctx = grn_ctx_pull_child(ctx_);
-        grn_obj *condition = build_condition(sub_ctx,
-                                             match_columns,
-                                             query.data(),
-                                             query.length());
+        grn::ChildCtxReleaser releaser(ctx_, sub_ctx);
         grn_obj *sub_result = nullptr;
-        if (condition) {
-          grn::UniqueObj unique_condition(sub_ctx, condition);
-          {
-            std::lock_guard<std::mutex> lock(mutex);
-            auto n_unique_sub_results = unique_sub_results.size();
-            if (n_unique_sub_results == 0) {
-              sub_result =
-                grn_table_create(ctx_,
-                                 nullptr, 0,
-                                 nullptr,
-                                 GRN_OBJ_TABLE_HASH_KEY|GRN_OBJ_WITH_SUBREC,
-                                 table_,
-                                 nullptr);
-            } else {
-              sub_result =
-                unique_sub_results[n_unique_sub_results - 1].release();
-              unique_sub_results.pop_back();
-            }
-          }
-          if (sub_result) {
-            grn_table_selector table_selector;
-            grn_table_selector_init(sub_ctx,
-                                    &table_selector,
-                                    table_,
-                                    condition,
-                                    GRN_OP_OR);
-            grn_table_selector_set_min_id(sub_ctx,
-                                          &table_selector,
-                                          min_id);
-            grn_table_selector_set_weight_factor(ctx_,
-                                                 &table_selector,
-                                                 get_weight_factor());
-            grn_table_selector_select(sub_ctx,
-                                      &table_selector,
-                                      sub_result);
-            grn_table_selector_fin(sub_ctx, &table_selector);
-            {
-              std::lock_guard<std::mutex> lock(mutex);
-              unique_sub_results.emplace_back(ctx_, sub_result);
-            }
+        {
+          std::lock_guard<std::mutex> lock(mutex);
+          auto n_unique_sub_results = unique_sub_results.size();
+          if (n_unique_sub_results < 2) {
+            sub_result =
+              grn_table_create(ctx_,
+                               nullptr, 0,
+                               nullptr,
+                               GRN_OBJ_TABLE_HASH_KEY|GRN_OBJ_WITH_SUBREC,
+                               table_,
+                               nullptr);
+          } else {
+            sub_result =
+              unique_sub_results[n_unique_sub_results - 1].release();
+            unique_sub_results.pop_back();
+            clear_result_set(ctx_, sub_result);
           }
         }
-        grn_ctx_release_child(ctx_, sub_ctx);
+        if (!select_one(sub_ctx,
+                        match_columns,
+                        query.data(),
+                        query.length(),
+                        GRN_OP_OR,
+                        min_id,
+                        &sub_result)) {
+          return ::arrow::Status::OK();
+        }
+        auto merge_future = merge_pool->Submit(merge, sub_result);
+        if (!merge_future.ok()) {
+          std::lock_guard<std::mutex> lock(mutex);
+          if (!grnarrow::check(ctx_,
+                               merge_future,
+                               "%s failed to submit a merge job",
+                               tag_)) {
+            return merge_future.status();
+          }
+        }
+        {
+          std::lock_guard<std::mutex> lock(mutex);
+          merge_futures.push_back(*merge_future);
+        }
         return ::arrow::Status::OK();
       };
 
       {
-        std::vector<::arrow::Future<>> futures;
+        std::vector<::arrow::Future<>> select_futures;
         auto n_sub_match_columns = GRN_PTR_VECTOR_SIZE(&sub_match_columns_);
         for (size_t i = 0; i < n_sub_match_columns; ++i) {
           grn_obj *sub_match_column = GRN_PTR_VALUE_AT(&sub_match_columns_, i);
           for (const auto &expanded_query : expanded_queries_) {
-            auto future = pool->Submit(select,
-                                       sub_match_column,
-                                       expanded_query);
-            if (!future.ok()) {
+            auto select_future = pool->Submit(select,
+                                              sub_match_column,
+                                              expanded_query);
+            if (!select_future.ok()) {
               std::lock_guard<std::mutex> lock(mutex);
               if (!grnarrow::check(ctx_,
-                                   future,
-                                   "%s failed to submit a job",
+                                   select_future,
+                                   "%s failed to submit a select job",
                                    tag_)) {
                 break;
               }
             }
-            futures.push_back(*future);
+            select_futures.push_back(*select_future);
           }
           if (ctx_->rc != GRN_SUCCESS) {
             break;
           }
         }
         auto status = ::arrow::Status::OK();
-        for (auto& future : futures) {
-          status &= future.status();
+        for (auto& select_future : select_futures) {
+          status &= select_future.status();
         }
         if (!grnarrow::check(ctx_,
                              status,
-                             "%s failed to complete a job",
+                             "%s failed to complete select jobs",
+                             tag_)) {
+          return false;
+        }
+        for (auto& merge_future : merge_futures) {
+          status &= merge_future.status();
+        }
+        if (!grnarrow::check(ctx_,
+                             status,
+                             "%s failed to complete merge jobs",
                              tag_)) {
           return false;
         }
       }
-      if (ctx_->rc != GRN_SUCCESS) {
-        return false;
-      }
-      auto sub_result = unique_sub_results[0].get();
-      auto n_sub_results = unique_sub_results.size();
-      for (size_t i = 1; i < n_sub_results; ++i) {
+      if (ctx_->rc == GRN_SUCCESS && !unique_sub_results.empty()) {
         grn_table_setoperation(ctx_,
-                               sub_result,
-                               unique_sub_results[i].get(),
-                               sub_result,
-                               GRN_OP_OR);
-        if (ctx_->rc != GRN_SUCCESS) {
-          break;
-        }
-      }
-      if (ctx_->rc == GRN_SUCCESS) {
-        grn_table_setoperation(ctx_, res_, sub_result, res_, op_);
-      }
+                               res_,
+                               unique_sub_results[0].get(),
+                               res_,
+                               op_);
+       }
       return ctx_->rc == GRN_SUCCESS;
     }
 #endif
@@ -744,70 +911,49 @@ namespace {
     select_sequential() {
       grn_id min_id = get_min_id();
       grn::UniqueObj unique_sub_result(ctx_, nullptr);
-      for (const auto &expanded_query : expanded_queries_) {
-        grn_obj *condition = build_condition(ctx_,
-                                             match_columns_,
-                                             expanded_query.data(),
-                                             expanded_query.length());
-        if (!condition) {
-          return false;
+      grn::UniqueObj unique_working_sub_result(ctx_, nullptr);
+      auto n_sub_match_columns = GRN_PTR_VECTOR_SIZE(&sub_match_columns_);
+      for (size_t i = 0; i < n_sub_match_columns; ++i) {
+        grn_obj *sub_match_column = GRN_PTR_VALUE_AT(&sub_match_columns_, i);
+        for (const auto &expanded_query : expanded_queries_) {
+          auto sub_result = unique_working_sub_result.get();
+          if (!select_one(ctx_,
+                          sub_match_column,
+                          expanded_query.data(),
+                          expanded_query.length(),
+                          GRN_OP_OR,
+                          min_id,
+                          &sub_result)) {
+            return false;
+          }
+          if (!unique_sub_result.get()) {
+            unique_sub_result.reset(sub_result);
+          } else {
+            grn_table_setoperation(ctx_,
+                                   unique_sub_result.get(),
+                                   sub_result,
+                                   unique_sub_result.get(),
+                                   GRN_OP_OR);
+            if (!unique_working_sub_result.get()) {
+              unique_working_sub_result.reset(sub_result);
+            }
+            clear_result_set(ctx_, unique_working_sub_result.get());
+          }
         }
-        grn::UniqueObj unique_condition(ctx_, condition);
-        grn_table_selector table_selector;
-        grn_table_selector_init(ctx_,
-                                &table_selector,
-                                table_,
-                                condition,
-                                GRN_OP_OR);
-        grn_table_selector_set_min_id(ctx_,
-                                      &table_selector,
-                                      min_id);
-        grn_table_selector_set_weight_factor(ctx_,
-                                             &table_selector,
-                                             get_weight_factor());
-        auto sub_result = grn_table_selector_select(ctx_,
-                                                    &table_selector,
-                                                    unique_sub_result.get());
-        grn_table_selector_fin(ctx_, &table_selector);
-        if (ctx_->rc != GRN_SUCCESS) {
-          return false;
-        }
-        unique_sub_result.reset(sub_result);
       }
       grn_table_setoperation(ctx_, res_, unique_sub_result.get(), res_, op_);
       return true;
     }
 
-    grn_obj *
-    build_condition(grn_ctx *ctx,
-                    grn_obj *match_columns,
-                    const char *query,
-                    size_t query_length) {
-      grn_obj *condition;
-      grn_obj *dummy_variable;
-      GRN_EXPR_CREATE_FOR_QUERY(ctx,
-                                table_,
-                                condition,
-                                dummy_variable);
-      if (!condition) {
-        return nullptr;
-      }
-      grn_expr_parse(ctx,
-                     condition,
-                     query,
-                     query_length,
-                     match_columns,
-                     default_mode_,
-                     default_operator_,
-                     flags_);
-      if (ctx->rc != GRN_SUCCESS) {
-        grn_obj_close(ctx, condition);
-        return nullptr;
-      }
-      if (query_options_) {
-        grn_expr_set_query_options(ctx_, condition, query_options_);
-      }
-      return condition;
+    void
+    clear_result_set(grn_ctx *ctx, grn_obj *result_set) {
+      // This is faster than grn_hash_truncate() or grn_table_create().
+      GRN_HASH_EACH_BEGIN(ctx,
+                          reinterpret_cast<grn_hash *>(result_set),
+                          cursor,
+                          id) {
+        grn_hash_cursor_delete(ctx, cursor, nullptr);
+      } GRN_HASH_EACH_END(ctx, cursor);
     }
 
     std::vector<std::string> expanded_queries_;
