@@ -32,7 +32,7 @@
 #include <stdio.h>
 
 static double grn_table_select_enough_filtered_ratio = 0.01;
-static int grn_table_select_max_n_enough_filtered_records = 1000;
+static int64_t grn_table_select_max_n_enough_filtered_records = 1000;
 static bool grn_table_select_min_id_skip_enable = true;
 static bool grn_query_log_show_condition = true;
 
@@ -113,6 +113,10 @@ grn_table_selector_init(grn_ctx *ctx,
   table_selector->use_sequential_scan = false;
   table_selector->query_options = grn_expr_get_query_options(ctx, expr);
   table_selector->weight_factor = 1.0;
+  table_selector->enough_filtered_ratio =
+    grn_table_select_enough_filtered_ratio;
+  table_selector->max_n_enough_filtered_records =
+    grn_table_select_max_n_enough_filtered_records;
   grn_table_selector_data data = {0};
   table_selector->data = data;
 }
@@ -203,6 +207,42 @@ grn_table_selector_set_weight_factor(grn_ctx *ctx,
   GRN_API_RETURN(ctx->rc);
 }
 
+double
+grn_table_selector_get_enough_filtered_ratio(grn_ctx *ctx,
+                                             grn_table_selector *table_selector)
+{
+  return table_selector->enough_filtered_ratio;
+}
+
+grn_rc
+grn_table_selector_set_enough_filtered_ratio(grn_ctx *ctx,
+                                             grn_table_selector *table_selector,
+                                             double ratio)
+{
+  GRN_API_ENTER;
+  table_selector->enough_filtered_ratio = ratio;
+  GRN_API_RETURN(ctx->rc);
+}
+
+int64_t
+grn_table_selector_get_max_n_enough_filtered_records(
+  grn_ctx *ctx,
+  grn_table_selector *table_selector)
+{
+  return table_selector->max_n_enough_filtered_records;
+}
+
+grn_rc
+grn_table_selector_set_max_n_enough_filtered_records(
+  grn_ctx *ctx,
+  grn_table_selector *table_selector,
+  int64_t n)
+{
+  GRN_API_ENTER;
+  table_selector->max_n_enough_filtered_records = n;
+  GRN_API_RETURN(ctx->rc);
+}
+
 static grn_obj *
 selector_create_result_set(grn_ctx *ctx,
                            grn_table_selector *table_selector)
@@ -424,24 +464,25 @@ select_index_use_sequential_search(grn_ctx *ctx,
     filtered_ratio = (double)n_filtered_records / (double)n_records;
   }
 
-  if (filtered_ratio >= grn_table_select_enough_filtered_ratio) {
+  if (filtered_ratio >= table_selector->enough_filtered_ratio) {
     return false;
   }
 
-  if (n_filtered_records > grn_table_select_max_n_enough_filtered_records) {
+  if (n_filtered_records > table_selector->max_n_enough_filtered_records) {
     return false;
   }
 
   grn_obj reason;
   GRN_TEXT_INIT(&reason, 0);
   grn_text_printf(ctx, &reason,
-                  "enough filtered: %.2f%%(%d/%d) < %.2f%% && %d <= %d",
+                  "enough filtered: "
+                  "%.2f%%(%d/%d) < %.2f%% && %d <= %" GRN_FMT_INT64D,
                   filtered_ratio * 100,
                   n_filtered_records,
                   n_records,
-                  grn_table_select_enough_filtered_ratio * 100,
+                  table_selector->enough_filtered_ratio * 100,
                   n_filtered_records,
-                  grn_table_select_max_n_enough_filtered_records);
+                  table_selector->max_n_enough_filtered_records);
   GRN_TEXT_PUTC(ctx, &reason, '\0');
   select_index_not_used_report(ctx,
                                tag,
@@ -545,9 +586,6 @@ select_index_equal(grn_ctx *ctx,
       if (tid == GRN_ID_NIL) {
         rc = GRN_SUCCESS;
       } else {
-        uint32_t sid = GRN_UINT32_VALUE_AT(&(si->sections), 0);
-        float weight = GRN_FLOAT32_VALUE_AT(&(si->weights), 0);
-        weight *= table_selector->weight_factor * si->weight_factor;
         grn_obj *index_cursor = grn_index_cursor_open(ctx,
                                                       NULL,
                                                       index,
@@ -557,7 +595,17 @@ select_index_equal(grn_ctx *ctx,
         if (index_cursor) {
           rc = grn_index_cursor_set_term_id(ctx, index_cursor, tid);
           if (rc == GRN_SUCCESS) {
-            rc = grn_index_cursor_set_section_id(ctx, index_cursor, sid);
+            grn_search_optarg *options = &(table_selector->data.search_options);
+            if (options->weight_vector_float) {
+              rc = grn_index_cursor_set_scales(ctx,
+                                               index_cursor,
+                                               options->weight_vector_float,
+                                               options->vector_size);
+            } else {
+              rc = grn_index_cursor_set_scale(ctx,
+                                              index_cursor,
+                                              options->weight_float);
+            }
           }
           if (si->position.specified) {
             if (rc == GRN_SUCCESS) {
@@ -571,7 +619,7 @@ select_index_equal(grn_ctx *ctx,
                                                  (grn_hash *)result_set,
                                                  index_cursor,
                                                  1,
-                                                 weight,
+                                                 1.0,
                                                  logical_op);
           }
           grn_obj_close(ctx, index_cursor);
@@ -671,17 +719,23 @@ select_index_not_equal(grn_ctx *ctx,
         grn_ii *ii = (grn_ii *)index;
         grn_ii_cursor *ii_cursor;
 
-        uint32_t sid = GRN_UINT32_VALUE_AT(&(si->sections), 0);
         ii_cursor = grn_ii_cursor_open(ctx, ii, tid,
                                        GRN_ID_NIL, GRN_ID_MAX,
                                        ii->n_elements, 0);
         if (ii_cursor) {
+          grn_search_optarg *options = &(table_selector->data.search_options);
+          if (options->weight_vector_float) {
+            rc = grn_ii_cursor_set_scales(ctx,
+                                          ii_cursor,
+                                          options->weight_vector_float,
+                                          options->vector_size);
+          } else {
+            rc = grn_ii_cursor_set_scale(ctx,
+                                         ii_cursor,
+                                         options->weight_float);
+          }
           grn_posting *posting;
           while ((posting = grn_ii_cursor_next(ctx, ii_cursor))) {
-            if (!(sid == 0 || posting->sid == sid)) {
-              continue;
-            }
-
             if (si->position.specified) {
               while ((posting = grn_ii_cursor_next_pos(ctx, ii_cursor))) {
                 if (posting->pos == si->position.start) {
@@ -809,12 +863,23 @@ select_index_fix_column_prefix(grn_ctx *ctx,
     GRN_OBJ_FIN(ctx, &query_inspected);
     goto exit;
   }
+  grn_search_optarg *options = &(table_selector->data.search_options);
+  if (options->weight_vector_float) {
+    grn_index_cursor_set_scales(ctx,
+                                index_cursor,
+                                options->weight_vector_float,
+                                options->vector_size);
+  } else {
+    grn_index_cursor_set_scale(ctx,
+                               index_cursor,
+                               options->weight_float);
+  }
   select_index_report(ctx, "[prefix]", index);
   grn_result_set_add_index_cursor(ctx,
                                   (grn_hash *)result_set,
                                   index_cursor,
                                   1,
-                                  1,
+                                  1.0,
                                   logical_op);
 exit :
   if (index_cursor) {
@@ -1119,6 +1184,7 @@ select_index_range_id(grn_ctx *ctx,
         grn_posting_internal posting = {0};
 
         posting.weight_float = weight;
+        posting.scale = 1.0;
         while ((posting.rid = grn_table_cursor_next(ctx, cursor))) {
           if (posting.rid < min) {
             continue;
@@ -1295,9 +1361,6 @@ select_index_range_column(grn_ctx *ctx,
                                    min, min_size, max, max_size,
                                    offset, limit, flags);
     if (cursor) {
-      uint32_t sid = GRN_UINT32_VALUE_AT(&(si->sections), 0);
-      float weight = GRN_FLOAT32_VALUE_AT(&(si->weights), 0);
-      weight *= table_selector->weight_factor * si->weight_factor;
       grn_obj *index_cursor = grn_index_cursor_open(ctx,
                                                     cursor,
                                                     index,
@@ -1305,7 +1368,17 @@ select_index_range_column(grn_ctx *ctx,
                                                     GRN_ID_MAX,
                                                     0);
       if (index_cursor) {
-        grn_index_cursor_set_section_id(ctx, index_cursor, sid);
+        grn_search_optarg *options = &(table_selector->data.search_options);
+        if (options->weight_vector_float) {
+          grn_index_cursor_set_scales(ctx,
+                                      index_cursor,
+                                      options->weight_vector_float,
+                                      options->vector_size);
+        } else {
+          grn_index_cursor_set_scale(ctx,
+                                     index_cursor,
+                                     options->weight_float);
+        }
         if (si->position.specified) {
           grn_index_cursor_set_start_position(ctx,
                                               index_cursor,
@@ -1315,7 +1388,7 @@ select_index_range_column(grn_ctx *ctx,
                                              (grn_hash *)result_set,
                                              index_cursor,
                                              1,
-                                             weight,
+                                             1.0,
                                              logical_op);
         grn_obj_close(ctx, index_cursor);
       }
